#!perl

use strict;
use warnings FATAL => 'all';

use lib qw{lib t/lib};

use Carp;
use Carp::Assert;
use Carp::Assert::More;
use Cwd;
use Data::Dumper;
use English qw{-no_match_vars};
use Env;
use File::Copy::Recursive qw{dircopy};
use File::Find::Rule;
use Test::More;
use Test::BDD::Cucumber::StepFile;
use Method::Signatures::Simple;
use Path::Class;
use File::Temp qw{tempdir tempfile};
use Tails::IUK::Utils qw{get_temp_dir run_as_root};
use Test::SslCertificates (
    qw{generate_ssl_privkey generate_self_signed_ssl_cert},
    qw{generate_ssl_req generate_ssl_cert populate_ssl_template}
);
use Test::Util;
use Test::WebServer::Static::SSL;

my $bindir = file(__FILE__)->dir->parent->parent->parent->subdir('bin')->absolute;
use Env qw{@PATH};
unshift @PATH, $bindir;

my $t_dir  = file(__FILE__)->dir->parent->parent->parent->subdir('t')->absolute;
my $pristine_dev_gnupg_homedir = file($t_dir, 'data', 'dev_gnupg_homedir');

my $update_description_file_relative_path = "update/v1/Tails/0.11/s390x/stable/updates.yml";
my $update_description_sig_relative_path  = "$update_description_file_relative_path".".pgp";
my $pristine_webroot = file($t_dir, qw{data webroot});

$ENV{HARNESS_ACTIVE} = 1;

Given qr{^a usable temporary directory$}, func($c) {
    my $tempdir = $c->{stash}->{scenario}->{tempdir} = get_temp_dir();
    ok(-d $tempdir);
    my $destdir = $c->{stash}->{scenario}->{destdir} = dir($tempdir, 'destdir');
    $destdir->mkpath;
    ok(-d $destdir);
};

Given qr{^a trusted OpenPGP signing key pair$}, func($c) {
    my $pristine_gnupg_homedir;
    my $name;

    $name = 'dev_gnupg_homedir';
    $pristine_gnupg_homedir = $pristine_dev_gnupg_homedir;
    assert(-d $pristine_gnupg_homedir);

    my $gnupg_homedir
        = $c->{stash}->{scenario}->{$name}
        = $c->{stash}->{scenario}->{trusted_gnupg_homedir}
        = dir($c->{stash}->{scenario}->{tempdir}, "$name")->absolute;

    dircopy($pristine_gnupg_homedir, $gnupg_homedir);
    assert(-d $gnupg_homedir);
    assert(-e file($gnupg_homedir, $_)) for qw{pubring.gpg secring.gpg};
};

Given qr{^a running Tails$}, func($c) {
    my $os_release_file
        = $c->{stash}->{scenario}->{os_release_file}
        = file($c->{stash}->{scenario}->{tempdir}, 'os_release_file');
    my $fh = $os_release_file->openw;
    my %os_release = (
        TAILS_PRODUCT_NAME => "Tails",
        TAILS_VERSION_ID   => "0.11",
        TAILS_CHANNEL      => "stable",
    );
    while (my ($key, $value) = each %os_release) {
        say $fh sprintf('%s="%s"', $key, $value);
    }
    close $fh;

    assert(-e $c->{stash}->{scenario}->{os_release_file});

    $c->{stash}->{scenario}->{liveos_mountpoint}
        = dir($c->{stash}->{scenario}->{tempdir}, 'live', 'image');
};

Given qr{^a Live filesystem$}, func($c) {
    $c->{stash}->{scenario}->{liveos_backing_file}
        = file($c->{stash}->{scenario}->{tempdir}, 'system.dev');
    Test::Util::prepare_live_image_filesystem(
        $c->{stash}->{scenario}->{liveos_backing_file},
        $c->{stash}->{scenario}->{liveos_mountpoint},
    );
};

Given qr{^a HTTPS server with a valid SSL certificate$}, func($c) {
    $c->{stash}->{scenario}->{server}->{https_port} = 40000 + int(rand(10000));
    ok(defined $c->{stash}->{scenario}->{server}->{https_port});

    my $ca_cert
        = $c->{stash}->{scenario}->{ca_cert}
        = file($c->{stash}->{scenario}->{tempdir}, "ca_cert")->stringify;
    my $ca_privkey
        = $c->{stash}->{scenario}->{ca_privkey}
        = file($c->{stash}->{scenario}->{tempdir}, "ca_privkey")->stringify;
    my $ca_template
        = file($c->{stash}->{scenario}->{tempdir}, "ca_template")->stringify;

    populate_ssl_template({ outfile => $ca_template, ca => 1 });
    assert(-e $ca_template);
    generate_ssl_privkey({ outfile => $ca_privkey });
    assert(-e $ca_privkey);
    generate_self_signed_ssl_cert({
        outfile => $ca_cert, privkey => $ca_privkey, template => $ca_template,
    });
    assert(-e $ca_cert);

    my $ssl_cert
        = $c->{stash}->{scenario}->{ssl_cert}
        = file($c->{stash}->{scenario}->{tempdir}, "ssl_cert")->stringify;
    my $ssl_privkey
        = $c->{stash}->{scenario}->{ssl_privkey}
        = file($c->{stash}->{scenario}->{tempdir}, "ssl_privkey")->stringify;
    my $ssl_template
        = file($c->{stash}->{scenario}->{tempdir}, "ssl_template")->stringify;
    my $ssl_req
        = file($c->{stash}->{scenario}->{tempdir}, "ssl_req")->stringify;

    populate_ssl_template({ outfile => $ssl_template, ca => 1 });
    assert(-e $ssl_template);
    generate_ssl_privkey({ outfile => $ssl_privkey });
    assert(-e $ssl_privkey);

    generate_ssl_req({ privkey => $ssl_privkey, outfile => $ssl_req, template => $ssl_template });
    assert(-e $ssl_req);

    generate_ssl_cert({
        req => $ssl_req,
        outfile => $ssl_cert,
        ca_cert => $c->{stash}->{scenario}->{ca_cert},
        ca_privkey => $c->{stash}->{scenario}->{ca_privkey},
        template   => $ssl_template,
    });
    assert(-e $ssl_cert);

    my $webroot
        = $c->{stash}->{scenario}->{webroot}
        = dir($c->{stash}->{scenario}->{tempdir}, 'webroot');
    dircopy($pristine_webroot, $webroot);
    assert(-d $webroot->stringify);

    my $port = $c->{stash}->{scenario}->{server}->{https_port};
    my $ca_file = $c->{stash}->{scenario}->{ca_cert};
    my $s = Test::WebServer::Static::SSL->new(
        {
            webroot => $c->{stash}->{scenario}->{webroot},
            cert    => $c->{stash}->{scenario}->{ssl_cert},
            key     => $c->{stash}->{scenario}->{ssl_privkey},
            ca      => $ca_file,
        },
        $port
    );
    is($s->port(), $port, "Constructor set port correctly");
    my $pid = $c->{stash}->{scenario}->{server}->{https_pid} = $s->background();
    like($pid, '/^-?\d+$/', 'PID is numeric');
};

func generate_update_description($output, $name, $version, $target, $channel, $gnupg_homedir, $port, $has_incremental_update, $has_full_update) {
    $has_incremental_update //= 0;
    $has_full_update        //= 0;
    my $desc = file($output);
    my $sig  = file($output.".pgp");

    $desc->dir->mkpath;
    my $desc_h = $desc->openw;
    print $desc_h Test::Util::update_description_header($name, $version, $target, $channel);
    if ($has_incremental_update or $has_full_update) {
        print $desc_h <<EOF
updates:
  - version: 0.12.1
    type: minor
    details-url: https://tails.boum.org/news/version_0.12.1/
    update-paths:

EOF
;
    }
    if ($has_incremental_update) {
        print $desc_h <<EOF
      - type: incremental
        target-files:
          - url: https://127.0.0.1:$port/tails/stable/iuk/Tails_s390x_0.12_to_0.12.1.iuk
            size: 5120
            sha256: cfc468e0451de01c8ab8b7ce388f799aaf979f645e957f22ba951268472e519b
EOF
;
    }
    if ($has_full_update) {
        print $desc_h <<EOF
      - type: full
        target-files:
          - url: https://127.0.0.1:$port/tails/stable/tails-s390x-0.12.1/Tails-s390x-0.12.1.iso
            size: 762123456
            sha256: a38c8b6566946170cb4ac806a36f20ddf0758a21b178b6f74c8268bbd7ecf8ab
EOF
;
    }
    close $desc_h;
    assert(-e $desc);

    system(
        qw{gpg --batch --quiet},
        qw{--armor --detach-sign},
        '--homedir', $gnupg_homedir,
        '--output',  $sig,
        $desc,
    );
    assert(-e $sig);
}

Given qr{^no update is available$}, func($c) {
    my $desc = file(
        $c->{stash}->{scenario}->{webroot},
        $update_description_file_relative_path,
    );
    my $sig  = file(
        $c->{stash}->{scenario}->{webroot},
        $update_description_sig_relative_path,
    );
    generate_update_description(
        $desc,
        "Tails", "0.11", "s390x", "stable",
        $c->{stash}->{scenario}->{dev_gnupg_homedir},
        $c->{stash}->{scenario}->{server}->{https_port},
    );
};

Given qr{^it is not known whether an update is available$}, func($c) {
    1;
};

Given qr{^an incremental update is available$}, func ($c) {
    my $desc = file(
        $c->{stash}->{scenario}->{webroot},
        $update_description_file_relative_path,
    );
    my $sig  = file(
        $c->{stash}->{scenario}->{webroot},
        $update_description_sig_relative_path,
    );
    generate_update_description(
        $desc,
        "Tails", "0.11", "s390x", "stable",
        $c->{stash}->{scenario}->{dev_gnupg_homedir},
        $c->{stash}->{scenario}->{server}->{https_port},
        "has incremental update", 0
    );
};

Given qr{^no incremental update is available, but a full update is$}, func ($c) {
    my $desc = file(
        $c->{stash}->{scenario}->{webroot},
        $update_description_file_relative_path,
    );
    my $sig  = file(
        $c->{stash}->{scenario}->{webroot},
        $update_description_sig_relative_path,
    );
    generate_update_description(
        $desc,
        "Tails", "0.11", "s390x", "stable",
        $c->{stash}->{scenario}->{dev_gnupg_homedir},
        $c->{stash}->{scenario}->{server}->{https_port},
        0, "has full update"
    );
};

Given qr{^a target file does not exist$}, func($c) {
    my @iuks = File::Find::Rule->file()
        ->name( '*.iuk' ) ->in($c->{stash}->{scenario}->{webroot});
    unlink for (@iuks);
};

Given qr{^a target file is corrupted$}, func($c) {
    my @iuks = File::Find::Rule->file()
        ->name( '*.iuk' ) ->in($c->{stash}->{scenario}->{webroot});
    foreach my $iuk (@iuks) {
        my $fh = file($iuk)->openw;
        say $fh "this is a corrupted IUK";
        close $fh;
    }
};

When qr{^I run tails-update-frontend(| in batch mode)$}, func($c) {
    my $batch = defined $c->matches->[0] && length defined $c->matches->[0];
    my $cmdline = sprintf("%s " .
            "--override-baseurl 'https://127.0.0.1:%s' " .
            "--override-os-release-file '%s' " .
            "--override-build-target '%s' " .
            "--override-trusted-gnupg-homedir '%s' ".
            "--liveos-mountpoint '%s'",
        file($bindir, "tails-update-frontend"),
        $c->{stash}->{scenario}->{server}->{https_port},
        $c->{stash}->{scenario}->{os_release_file},
        's390x',
        $c->{stash}->{scenario}->{trusted_gnupg_homedir},
        $c->{stash}->{scenario}->{liveos_mountpoint},
    );
    $cmdline .= " --batch" if $batch;
    $ENV{HTTPS_CA_FILE} = $c->{stash}->{scenario}->{ca_cert};
    $c->{stash}->{scenario}->{output} = `$cmdline 2>&1`;
    $c->{stash}->{scenario}->{exit_code} = ${^CHILD_ERROR_NATIVE};
};

Then qr{^it should succeed$}, func($c) {
    Test::Util::kill_httpd($c);

    ok(defined $c->{stash}->{scenario}->{exit_code})
        and
    is($c->{stash}->{scenario}->{exit_code}, 0);

    if (defined $c->{stash}->{scenario}->{exit_code}
            && $c->{stash}->{scenario}->{exit_code} != 0
            && exists $c->{stash}->{scenario}->{output}
            && defined $c->{stash}->{scenario}->{output}) {
        warn $c->{stash}->{scenario}->{output};
    }
};

Then qr{^it should fail$}, func($c) {
    Test::Util::kill_httpd($c);

    ok(defined $c->{stash}->{scenario}->{exit_code})
        and
    isnt($c->{stash}->{scenario}->{exit_code}, 0);

    if (exists $c->{stash}->{scenario}->{output}
            && defined $c->{stash}->{scenario}->{output}) {
        warn $c->{stash}->{scenario}->{output};
    }
};

Then qr{^I should not be told anything}, func($c) {
    is($c->{stash}->{scenario}->{output}, "");
};

Then qr{^I should be pointed to the documentation about update-description file retrieval error$}, func($c) {
    like(
        $c->{stash}->{scenario}->{output},
        qr{/usr/share/doc/tails/website/doc/upgrade/error/get_update_description}
    );
};

Then qr{^I should be pointed to the documentation about target file retrieval error$}, func($c) {
    like(
        $c->{stash}->{scenario}->{output},
        qr{/usr/share/doc/tails/website/doc/upgrade/error/get_target_file}
    );
};

Then qr{^I should be told "([^"]+)"$}, func($c) {
    my $expected_err = $c->matches->[0];
    like($c->{stash}->{scenario}->{output}, qr{$expected_err});
};

Then qr{^I should be proposed to install this update$}, func($c) {
    like(
        $c->{stash}->{scenario}->{output},
        qr{You should update to Tails 0[.]12[.]1.*https://tails.boum.org/news/version_0[.]12[.]1/}
    );
};

Then qr{^the web page about the new release should be opened$}, func($c) {
    like(
        $c->{stash}->{scenario}->{output},
        qr{Opening https://tails[.]boum[.]org/news/version_0[.]12[.]1/ in web browser}
    );
};

Then qr{^I should be asked to wait$}, func($c) {
    like(
        $c->{stash}->{scenario}->{output},
        qr{please wait}
    );
};

Then qr{^the network should be shutdown$}, func($c) {
    like(
        $c->{stash}->{scenario}->{output},
        qr{Shutting down network access}
    );
};

Then qr{^the downloaded IUK should be installed$}, func($c) {
    # one of the tarballs shipped by the test IUK contains a "placeholder" file
    ok(-e file($c->{stash}->{scenario}->{liveos_mountpoint}, 'placeholder'));
};

Then qr{^I should be proposed to halt the system$}, func($c) {
    like(
        $c->{stash}->{scenario}->{output},
        qr{You should reboot}
    );
};

Then qr{^the system should be halted$}, func($c) {
    like(
        $c->{stash}->{scenario}->{output},
        qr{Shutting down the system}
    );
};

Then qr{^the test environment should be cleaned up$}, func($c) {
    # Cleanup behind us... because this is the last step in every scenario.
    # Ugly, sure, but Test::BDD::Cucumber does not provide us with something
    # similar to Background, that would run *after* every scenario.
    run_as_root('umount', $c->{stash}->{scenario}->{liveos_mountpoint});
    ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to umount live image.");
};
