#!perl

use strict;
use warnings FATAL => 'all';

use lib qw{lib t/lib};

use Carp;
use Carp::Assert;
use Carp::Assert::More;
use Cwd;
use Data::Dumper;
use English qw{-no_match_vars};
use Path::Class;
use Test::More;
use Test::BDD::Cucumber::StepFile;
use Method::Signatures::Simple;

use File::Temp qw{tempdir tempfile};
use YAML::Any;

use Tails::IUK;
use Tails::IUK::Install;
use Tails::IUK::Utils qw{get_temp_dir make_archive_with_files_called make_iuk_with_files run_as_root};

use Test::Util;

my $bindir = file(__FILE__)->dir->parent->parent->parent->subdir('bin')->absolute;
use Env qw{@PATH};
unshift @PATH, $bindir;

Given qr{^a usable temporary directory$}, func($c) {
    my $dirname = get_temp_dir();
    $c->{stash}->{scenario}->{tempdir} = $dirname;
    ok(defined($dirname));
};

Given qr{^a usable (\d+)MB live image directory$}, func($c) {
    my $size = $c->matches->[0];
    my $live_image
        = $c->{stash}->{scenario}->{live_image}
        = dir($c->{stash}->{scenario}->{tempdir}, 'live', 'image');
    $live_image->mkpath;

    my $system_dev_backing_file
        = $c->{stash}->{scenario}->{system_dev_backing_file}
        = file($c->{stash}->{scenario}->{tempdir}, 'system.dev');

    Test::Util::prepare_live_image_filesystem(
        $system_dev_backing_file,
        $live_image,
        $size
    );
};

Given qr{^a "([^"]+)" IUK that contains (\d+) SquashFS$}, func($c) {
    my $iuk_filename    = $c->matches->[0];
    my $wanted_squashfs = $c->matches->[1];
    my $orig_cwd = getcwd;

    chdir $c->{stash}->{scenario}->{tempdir};
    my @files = ('placeholder');
    if ($wanted_squashfs) {
        push @files, map { file('live', "$_.squashfs") } (1..$wanted_squashfs);
    }
    my ($archive_fh, $archive_file) =
        tempfile(DIR => $c->{stash}->{scenario}->{tempdir});
    close $archive_fh;
    rename $archive_file, "$archive_file.tar";
    $archive_file = file("$archive_file.tar");
    make_archive_with_files_called($archive_file,
                                   content => 'test', filenames => \@files);

    chdir $archive_file->dir;
    ok(make_iuk_with_files(
        file($c->{stash}->{scenario}->{tempdir}, $iuk_filename),
        $c->{stash}->{scenario}->{tempdir},
        $archive_file->basename,
    ));
    chdir $orig_cwd;
};

Given qr{^a "([^"]+)" archive that ships(?:| the (\d+)MB) file "([^"]+)"(?: with content "([^"]*)")?$}, func($c) {
    my $archive_filename = $c->matches->[0];
    my $shipped_filesize = $c->matches->[1];
    my $shipped_filename = $c->matches->[2];
    my $content          = $c->matches->[3];

    my %args = ( filenames => [$shipped_filename]);
    if (defined $content) {
        $args{content} = $content;
    }
    elsif (defined $shipped_filesize) {
        $args{size} = $shipped_filesize;
    }
    else {
        $args{content} = "content";
    }

    my $archive = file($c->{stash}->{scenario}->{tempdir}, $archive_filename);
    make_archive_with_files_called($archive, %args);
    ok(-e $archive);
};

Given qr{^a "([^"]+)" IUK that contains no archive$}, func($c) {
    my $iuk_filename    = $c->matches->[0];

    ok(make_iuk_with_files(
        file($c->{stash}->{scenario}->{tempdir}, $iuk_filename),
        $c->{stash}->{scenario}->{tempdir},
    ));
};

Given qr{^a "([^"]+)" IUK whose format version is not supported$}, func($c) {
    my $iuk_filename    = $c->matches->[0];

    my $iuk = Archive::Tar->new();
    chdir $c->{stash}->{scenario}->{tempdir};
    my $fh = file('FORMAT')->openw;
    print $fh 42;
    close $fh;
    $iuk->add_files('FORMAT');
    chdir '/';
    ok($iuk->write(file($c->{stash}->{scenario}->{tempdir}, $iuk_filename)));
};

Given qr{^a "([^"]+)" IUK whose format version cannot be determined$}, func($c) {
    my $iuk_filename    = $c->matches->[0];

    my $iuk = Archive::Tar->new();
    chdir $c->{stash}->{scenario}->{tempdir};
    my $fh = file('FORMAT')->openw;
    print $fh "abc";
    close $fh;
    $iuk->add_files('FORMAT');
    chdir '/';
    ok($iuk->write(file($c->{stash}->{scenario}->{tempdir}, $iuk_filename)));
};

Given qr{^a "([^"]+)" IUK that has no FORMAT file$}, func($c) {
    my $iuk_filename    = $c->matches->[0];

    my $iuk = Archive::Tar->new();
    chdir $c->{stash}->{scenario}->{tempdir};
    my $fh = file('whatever.file')->openw;
    print $fh "abc";
    close $fh;
    $iuk->add_files('whatever.file');
    chdir '/';
    ok($iuk->write(file($c->{stash}->{scenario}->{tempdir}, $iuk_filename)));
};

Given qr{^a running Tails that has no IUK installed$}, func($c) {
    is(scalar(squashfs_in_live_image($c->{stash}->{scenario}->{live_image})), 1);
};

Given qr{^a "([^"]+)" IUK that contains archive[s]? "([^"]+)"(?: and "([^"]+)")?$}, func($c) {
    my $iuk_filename    = $c->matches->[0];
    my @included_archives = ($c->matches->[1]);
    push @included_archives, $c->matches->[2] if defined $c->matches->[2];

    ok(make_iuk_with_files(
        file($c->{stash}->{scenario}->{tempdir}, $iuk_filename),
        $c->{stash}->{scenario}->{tempdir},
        @included_archives,
    ));
};

Given qr{^a "([^"]+)" IUK that should delete files "([^"]*)"$}, func($c) {
    my $iuk_filename    = $c->matches->[0];
    my @delete_files = split /,[[:blank:]]+/, $c->matches->[1];
    my $orig_cwd = getcwd;

    chdir $c->{stash}->{scenario}->{tempdir};
    my $control = file($c->{stash}->{scenario}->{tempdir}, 'control.yml');
    my $control_fh = $control->openw;
    print $control_fh YAML::Any::Dump({ delete_files => \@delete_files });
    close $control_fh;

    ok(make_iuk_with_files(
        file($c->{stash}->{scenario}->{tempdir}, $iuk_filename),
        $c->{stash}->{scenario}->{tempdir},
        $control->basename,
    ));
    chdir $orig_cwd;
};

Given qr{^a live image that contains file "([^"]+)"$}, func($c) {
    my $file = file($c->{stash}->{scenario}->{live_image}, $c->matches->[0]);

    Test::Util::remount_loop_for_me_rw(
        $c->{stash}->{scenario}->{system_dev_backing_file},
        $c->{stash}->{scenario}->{live_image});

    $file->dir->mkpath;
    $file->touch;

    Test::Util::remount_loop_for_root_ro(
        $c->{stash}->{scenario}->{system_dev_backing_file},
        $c->{stash}->{scenario}->{live_image});

    ok(  Test::Util::has_mount_option($c->{stash}->{scenario}->{live_image}, 'ro'));
    ok(! Test::Util::has_mount_option($c->{stash}->{scenario}->{live_image}, 'rw'));
};

When qr{^I (?:attempt to |)install the "([^"]+)" IUK$}, func($c) {
    my $iuk = $c->matches->[0];
    local $@;
    my $cmdline =
        file($bindir, "tails-install-iuk") .
        ' --liveos-mountpoint "' . $c->{stash}->{scenario}->{live_image} . '" "' .
        file($c->{stash}->{scenario}->{tempdir}, $iuk) . '"';
    $c->{stash}->{scenario}->{install_output} = `$cmdline 2>&1`;
    $c->{stash}->{scenario}->{install_exit_code} = ${^CHILD_ERROR_NATIVE};
};

Then qr{^it should fail$}, func($c) {
    ok(defined $c->{stash}->{scenario}->{install_exit_code})
        and
    isnt($c->{stash}->{scenario}->{install_exit_code}, 0);
};

Then qr{^it should succeed$}, func($c) {
    ok(defined $c->{stash}->{scenario}->{install_exit_code});
    is($c->{stash}->{scenario}->{install_exit_code}, 0);

    if (! (defined $c->{stash}->{scenario}->{install_exit_code}
               && $c->{stash}->{scenario}->{install_exit_code} == 0)
            && exists $c->{stash}->{scenario}->{install_output}
            && defined $c->{stash}->{scenario}->{install_output}) {
        warn $c->{stash}->{scenario}->{install_output};
    }
};

Then qr{^the temporary directory on the live image should be empty$}, func($c) {
    my $tempdir = $c->{stash}->{scenario}->{live_image}->subdir('tmp');
    ok(
        ! -d $tempdir || scalar($tempdir->children) == 0
    );
};

Then qr{^I should be told "([^"]+)"$}, func($c) {
    my $expected_err = $c->matches->[0];
    like($c->{stash}->{scenario}->{install_output}, qr{$expected_err});
};

func squashfs_in_live_image ($live_image) {
    grep {
        $_ =~ m{[.]squashfs \z}xms;
    } dir($live_image, 'live')->children
}

Then qr{^the live image should contain (\d+) SquashFS$}, func($c) {
    my $expected_squashfs = $c->matches->[0];
    is(
        scalar(squashfs_in_live_image($c->{stash}->{scenario}->{live_image})),
        $expected_squashfs
    );
};

Then qr{^the live image should (not |)contain file "([^"]+)"(?: with content "([^"]*)")?$}, func($c) {
    my $should_exist     = $c->matches->[0] ? 0 : 1;
    my $file             = $c->matches->[1];
    my $expected_content = $c->matches->[2];

    if ($should_exist) {
        $file = file($c->{stash}->{scenario}->{live_image}, $file);
        ok(-e $file);
        if (defined $expected_content) {
            is($file->slurp, $expected_content);
        }
    }
    else {
        ok(! -e file($c->{stash}->{scenario}->{live_image}, $file));
    }
};

Then qr{^the last line of the modules file should be "([^"]+)"$}, func($c) {
    my $expected_last_module = $c->matches->[0];

    my @lines = grep {
        ! m{\A[[:blank:]]*\z}
    } file($c->{stash}->{scenario}->{live_image}, 'live', 'Tails.module')
        ->slurp(chomp => 1);
    is($lines[-1], $expected_last_module);
};

Then qr{^the test environment should be cleaned up$}, func($c) {
    # Cleanup behind us... because this is the last step in every scenario.
    # Ugly, sure, but Test::BDD::Cucumber does not provide us with something
    # similar to Background, that would run *after* every scenario.
    run_as_root('umount', $c->{stash}->{scenario}->{live_image});
    ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to umount live image.");
};
