package Test::Util;

use strict;
use warnings FATAL => 'all';
use 5.10.1;

use Carp;
use English qw{-no_match_vars};
use Method::Signatures::Simple;
use Path::Class;
use Sys::Filesystem ();
use Tails::IUK::Utils qw{run_as_root};
use Test::More;

func kill_httpd ($c) {
    foreach my $pid_key (qw{http_pid https_pid}) {
        my $pid = $c->{stash}->{scenario}->{server}->{$pid_key};
        is(kill(9, $pid), 1, "Killed PID $pid" ) if defined $pid;
    }
}

func update_description_header ($product_name, $product_version, $build_target, $channel) {
    return <<EOF
product-name: $product_name
product-version: $product_version
channel: $channel
build-target: $build_target
EOF
;
}

# $mount may be a device or mountpoint
func mount_options ($mount) {
    my $fs = Sys::Filesystem->new();
    return split(/,/, $fs->options($mount));
}

# $mount may be a device or mountpoint
func has_mount_option ($mount, $option) {
    grep { $_ eq $option } mount_options($mount);
}

func remount_loop_for_root_ro ($backing_file, $mountpoint) {
    run_as_root('umount', $mountpoint);
    ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to umount '$mountpoint'.");

    run_as_root('mount', '-o', 'loop', $backing_file, $mountpoint);
    ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to mount '$mountpoint'.");

    run_as_root('mount', '-o', 'remount,ro', $mountpoint);
    ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to remount '$mountpoint' read-only.");
}

func remount_loop_for_me_rw ($backing_file, $mountpoint) {
    run_as_root('umount', $mountpoint);
    ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to umount '$mountpoint'.");

    my @gids = split(/ /, $GID);
    run_as_root('mount', '-o', "loop,rw,uid=$UID,gid=$gids[0]",
                $backing_file, $mountpoint);
    ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to mount '$mountpoint'.");
}

# $size is in MB
func prepare_live_image_filesystem ($backing_file, $mountpoint, $size) {
    $size //= 64;

    system(qw{mkdosfs -C -F 32}, $backing_file, $size * 2**10);
    ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to create backing file.");

    dir($mountpoint)->mkpath;

    my @gids = split(/ /, $GID);
    run_as_root(
        "mount", "-o", "loop,uid=$UID,gid=$gids[0]", '-t', 'vfat',
        $backing_file, $mountpoint
    );
    ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to loop-mount backing file.");

    dir($mountpoint, 'live')->mkpath;
    my $squashfs = file($mountpoint, 'live', 'filesystem.squashfs');
    $squashfs->touch;
    my $modules_file = file($mountpoint, 'live', 'Tails.module');
    my $modules_fh = $modules_file->openw;
    say $modules_fh 'filesystem.squashfs';
    close $modules_fh;

    remount_loop_for_root_ro($backing_file, $mountpoint);

    ok(-e $squashfs) and ok(-e $modules_file);
    ok(  has_mount_option($mountpoint, 'ro'));
    ok(! has_mount_option($mountpoint, 'rw'));
}

1;
